# ctypes wrapper for libqpoint

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import ctypes as ct
import numpy as np
import os

libqp = np.ctypeslib.load_library('libqpoint', os.path.dirname(__file__))

# **********************************************************************
# Types
# **********************************************************************

NDP = np.ctypeslib.ndpointer

quat_t = NDP(np.double, ndim=1, shape=(4,), flags=['A', 'C'])
quat_t_p = NDP(np.double, ndim=2, flags=['A', 'C'])
mueller_t = NDP(np.double, ndim=1, shape=(4,), flags=['A', 'C'])
mueller_t_p = NDP(np.double, ndim=2, flags=['A', 'C'])
vec3_t = NDP(np.double, ndim=1, shape=(3,), flags=['A', 'C'])
vec3_t_p = NDP(np.double, ndim=2, flags=['A', 'C'])

wquat_t = NDP(np.double, ndim=1, shape=(4,), flags=['A', 'C', 'W'])
wquat_t_p = NDP(np.double, ndim=2, flags=['A', 'C', 'W'])
wvec3_t = NDP(np.double, ndim=1, shape=(3,), flags=['A', 'C', 'W'])
wvec3_t_p = NDP(np.double, ndim=2, flags=['A', 'C', 'W'])

arr = NDP(np.double, ndim=1, flags=['A', 'C'])
arrf = NDP(np.uint8, ndim=1, flags=['A', 'C'])
warr = NDP(np.double, ndim=1, flags=['A', 'C', 'W'])
warri = NDP(int, ndim=1, flags=['A', 'C', 'W'])

arr2 = NDP(np.uintp, ndim=1, flags=['A', 'C'])
larr = NDP(int, ndim=1, flags=['A', 'C'])

QP_DO_ALWAYS = ct.c_int.in_dll(libqp, "QP_DO_ALWAYS").value
QP_DO_ONCE = ct.c_int.in_dll(libqp, "QP_DO_ONCE").value
QP_DO_NEVER = ct.c_int.in_dll(libqp, "QP_DO_NEVER").value


class qp_state_t(ct.Structure):
    _fields_ = [('update_rate', ct.c_double), ('ctime_last', ct.c_double)]


class qp_weather_t(ct.Structure):
    _fields_ = [
        ('temperature', ct.c_double),
        ('pressure', ct.c_double),
        ('humidity', ct.c_double),
        ('frequency', ct.c_double),
    ]


class qp_bulletina_entry_t(ct.Structure):
    _fields_ = [
        ('x', ct.c_float),
        ('y', ct.c_float),
        ('dut1', ct.c_float),
    ]


class qp_bulletina_t(ct.Structure):
    _fields_ = [
        ('entries', ct.POINTER(qp_bulletina_entry_t)),
        ('mjd_min', ct.c_int),
        ('mjd_max', ct.c_int),
    ]


class qp_memory_t(ct.Structure):
    _fields_ = [
        ('init', ct.c_int),
        ('state_daber', qp_state_t),
        ('state_lonlat', qp_state_t),
        ('state_wobble', qp_state_t),
        ('state_dut1', qp_state_t),
        ('state_erot', qp_state_t),
        ('state_npb', qp_state_t),
        ('state_aaber', qp_state_t),
        ('state_ref', qp_state_t),
        ('state_daber_inv', qp_state_t),
        ('state_lonlat_inv', qp_state_t),
        ('state_wobble_inv', qp_state_t),
        ('state_dut1_inv', qp_state_t),
        ('state_erot_inv', qp_state_t),
        ('state_npb_inv', qp_state_t),
        ('state_aaber_inv', qp_state_t),
        ('state_ref_inv', qp_state_t),
        ('weather', qp_weather_t),
        ('ref_delta', ct.c_double),
        ('q_ref', ct.c_double * 4),
        ('q_ref_inv', ct.c_double * 4),
        ('dut1', ct.c_double),
        ('q_lonlat', ct.c_double * 4),
        ('q_lonlat_inv', ct.c_double * 4),
        ('q_wobble', ct.c_double * 4),
        ('q_wobble_inv', ct.c_double * 4),
        ('q_npb', ct.c_double * 4),
        ('q_npb_inv', ct.c_double * 4),
        ('q_erot', ct.c_double * 4),
        ('q_erot_inv', ct.c_double * 4),
        ('q_gal', ct.c_double * 4),
        ('q_gal_inv', ct.c_double * 4),
        ('gal_init', ct.c_int),
        ('v_dipole', ct.c_double * 3),
        ('dipole_init', ct.c_int),
        ('beta_earth', ct.c_double * 3),
        ('beta_rot', ct.c_double * 3),
        ('bulletinA', qp_bulletina_t),
        ('accuracy', ct.c_int),
        ('mean_aber', ct.c_int),
        ('fast_math', ct.c_int),
        ('polconv', ct.c_int),
        ('pix_order', ct.c_int),
        ('interp_pix', ct.c_int),
        ('fast_pix', ct.c_int),
        ('error_missing', ct.c_int),
        ('nan_missing', ct.c_int),
        ('interp_missing', ct.c_int),
        ('num_threads', ct.c_int),
        ('thread_num', ct.c_int),
    ]


qp_memory_t_p = ct.POINTER(qp_memory_t)

QP_STRUCT_INIT = 1
QP_STRUCT_MALLOC = 2
QP_ARR_INIT_PTR = 4
QP_ARR_MALLOC_1D = 8
QP_ARR_MALLOC_2D = 16


class qp_det_t(ct.Structure):
    _fields_ = [
        ('init', ct.c_int),
        ('q_off', ct.c_double * 4),
        ('weight', ct.c_double),
        ('gain', ct.c_double),
        ('mueller', ct.c_double * 4),
        ('n', ct.c_size_t),
        ('tod_init', ct.c_int),
        ('tod', ct.POINTER(ct.c_double)),
        ('flag_init', ct.c_int),
        ('flag', ct.POINTER(ct.c_uint8)),
        ('weights_init', ct.c_int),
        ('weights', ct.POINTER(ct.c_double)),
    ]


qp_det_t_p = ct.POINTER(qp_det_t)


class qp_detarr_t(ct.Structure):
    _fields_ = [
        ('init', ct.c_int),
        ('n', ct.c_size_t),
        ('arr_init', ct.c_int),
        ('diff', ct.c_size_t),
        ('arr', qp_det_t_p),
    ]


qp_detarr_t_p = ct.POINTER(qp_detarr_t)


class qp_point_t(ct.Structure):
    _fields_ = [
        ('init', ct.c_int),
        ('n', ct.c_size_t),
        ('q_bore_init', ct.c_int),
        ('q_bore', ct.POINTER(ct.c_double * 4)),
        ('ctime_init', ct.c_int),
        ('ctime', ct.POINTER(ct.c_double)),
        ('q_hwp_init', ct.c_int),
        ('q_hwp', ct.POINTER(ct.c_double * 4)),
    ]


qp_point_t_p = ct.POINTER(qp_point_t)

qp_vec_mode = ct.c_uint
QP_VEC_NONE = 0
QP_VEC_TEMP = 1
QP_VEC_POL = 2
QP_VEC_VPOL = 3
QP_VEC_D1 = 4
QP_VEC_D1_POL = 5
QP_VEC_D2 = 6
QP_VEC_D2_POL = 7
vec_modes = {
    1: QP_VEC_TEMP,
    3: {True: QP_VEC_POL, False: QP_VEC_D1},
    4: QP_VEC_VPOL,
    6: QP_VEC_D2,
    9: QP_VEC_D1_POL,
    18: QP_VEC_D2_POL,
}

qp_proj_mode = ct.c_uint
QP_PROJ_NONE = 0
QP_PROJ_TEMP = 1
QP_PROJ_POL = 2
QP_PROJ_VPOL = 3
proj_modes = {1: QP_PROJ_TEMP, 6: QP_PROJ_POL, 10: QP_PROJ_VPOL}


class qp_map_t(ct.Structure):
    _fields_ = [
        ('init', ct.c_int),
        ('partial', ct.c_int),
        ('nside', ct.c_size_t),
        ('npix', ct.c_size_t),
        ('pixinfo_init', ct.c_int),
        ('pixinfo', ct.c_void_p),
        ('pixhash_init', ct.c_int),
        ('pixhash', ct.c_void_p),
        ('num_vec', ct.c_size_t),
        ('vec_mode', qp_vec_mode),
        ('vec1d_init', ct.c_int),
        ('vec1d', ct.POINTER(ct.c_double)),
        ('vec_init', ct.c_int),
        ('vec', ct.POINTER(ct.POINTER(ct.c_double))),
        ('num_proj', ct.c_size_t),
        ('proj_mode', qp_proj_mode),
        ('proj1d_init', ct.c_int),
        ('proj1d', ct.POINTER(ct.c_double)),
        ('proj_init', ct.c_int),
        ('proj', ct.POINTER(ct.POINTER(ct.c_double))),
    ]


qp_map_t_p = ct.POINTER(qp_map_t)


def pointer_2d(d):
    return (
        d.__array_interface__['data'][0] + np.arange(d.shape[0]) * d.strides[0]
    ).astype(np.uintp)


def as_ctypes(d):
    return np.ctypeslib.as_ctypes(d)


def setargs(fname, arg=None, res=None):
    func = getattr(libqp, fname)
    if arg is not None and not isinstance(arg, (tuple, list)):
        arg = (arg,)
    func.argtypes = arg
    func.restype = res


# **********************************************************************
# Functions
# **********************************************************************

setargs('qp_init_memory', res=qp_memory_t_p)
setargs('qp_free_memory', arg=qp_memory_t_p)
setargs('qp_print_memory', arg=qp_memory_t_p)

setargs('qp_reset_rates', arg=qp_memory_t_p)
setargs('qp_reset_inv_rates', arg=qp_memory_t_p)

setargs('qp_get_error_code', arg=qp_memory_t_p, res=ct.c_int)
setargs('qp_get_error_string', arg=qp_memory_t_p, res=ct.c_char_p)

setargs(
    'qp_azel2radec',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2radec',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azel2radecpa',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2radecpa',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_radec2azel',
    arg=(
        qp_memory_t_p,  # params
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # r/d/p/l/l/t/a/e/p
        ct.c_int,
    ),
)
setargs(
    'qp_azel2radec_hwp',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t/hwp
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2radec_hwp',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t/hwp
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azel2radecpa_hwp',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t/hwp
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2radecpa_hwp',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t/hwp
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azel2rasindec',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2rasindec',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azel2rasindec_hwp',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t/hwp
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2rasindec_hwp',
    arg=(
        qp_memory_t_p,  # params
        ct.c_double,
        ct.c_double,
        ct.c_double,  # offset
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t/hwp
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_azel2bore',
    arg=(
        qp_memory_t_p,  # params
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/r/l/l/t
        wquat_t_p,
        ct.c_int,
    ),
)
setargs(
    'qp_azelpsi2bore',
    arg=(
        qp_memory_t_p,  # params
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,
        arr,  # a/e/p/p/r/l/l/t
        wquat_t_p,
        ct.c_int,
    ),
)

setargs('qp_det_offsetn', arg=(arr, arr, arr, wquat_t_p, ct.c_int))
setargs(
    'qp_bore_offset', arg=(qp_memory_t_p, wquat_t_p, arr, arr, arr, ct.c_int, ct.c_int)
)
setargs('qp_hwp_quatn', arg=(arr, wquat_t_p, ct.c_int))
setargs('qp_gmstn', arg=(qp_memory_t_p, arr, warr, ct.c_int))
setargs('qp_lmstn', arg=(qp_memory_t_p, arr, arr, warr, ct.c_int))
setargs('qp_dipolen', arg=(qp_memory_t_p, arr, arr, arr, warr, ct.c_int))
setargs('qp_bore2dipole', arg=(qp_memory_t_p, quat_t, arr, quat_t_p, warr, ct.c_int))

setargs(
    'qp_bore2radec',
    arg=(qp_memory_t_p, quat_t, arr, quat_t_p, warr, warr, warr, warr, ct.c_int),
)
setargs(
    'qp_bore2radec_hwp',
    arg=(
        qp_memory_t_p,
        quat_t,
        arr,
        quat_t_p,
        quat_t_p,
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_bore2rasindec',
    arg=(qp_memory_t_p, quat_t, arr, quat_t_p, warr, warr, warr, warr, ct.c_int),
)
setargs(
    'qp_bore2rasindec_hwp',
    arg=(
        qp_memory_t_p,
        quat_t,
        arr,
        quat_t_p,
        quat_t_p,
        warr,
        warr,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_bore2radecpa',
    arg=(qp_memory_t_p, quat_t, arr, quat_t_p, warr, warr, warr, ct.c_int),
)
setargs(
    'qp_bore2radecpa_hwp',
    arg=(qp_memory_t_p, quat_t, arr, quat_t_p, quat_t_p, warr, warr, warr, ct.c_int),
)

setargs('qp_radecpa2quatn', arg=(qp_memory_t_p, arr, arr, arr, wquat_t_p, ct.c_int))
setargs('qp_quat2radecpan', arg=(qp_memory_t_p, quat_t_p, warr, warr, warr, ct.c_int))

setargs('qp_radec2pixn', arg=(qp_memory_t_p, arr, arr, ct.c_int, warri, ct.c_int))

setargs('qp_radec2gal_quatn', arg=(qp_memory_t_p, wquat_t_p, ct.c_int))
setargs('qp_gal2radec_quatn', arg=(qp_memory_t_p, wquat_t_p, ct.c_int))

setargs('qp_radec2galn', arg=(qp_memory_t_p, warr, warr, warr, warr, ct.c_int))
setargs('qp_gal2radecn', arg=(qp_memory_t_p, warr, warr, warr, warr, ct.c_int))
setargs('qp_radecpa2galn', arg=(qp_memory_t_p, warr, warr, warr, ct.c_int))
setargs('qp_gal2radecpan', arg=(qp_memory_t_p, warr, warr, warr, ct.c_int))
setargs(
    'qp_rotate_map', arg=(qp_memory_t_p, ct.c_int, arr2, ct.c_char, arr2, ct.c_char)
)

setargs(
    'qp_quat2pixn', arg=(qp_memory_t_p, quat_t_p, ct.c_int, warri, warr, warr, ct.c_int)
)
setargs(
    'qp_quat2pixpan', arg=(qp_memory_t_p, quat_t_p, ct.c_int, warri, warr, ct.c_int)
)
setargs(
    'qp_bore2pix',
    arg=(qp_memory_t_p, quat_t, arr, quat_t_p, ct.c_int, warri, warr, warr, ct.c_int),
)
setargs(
    'qp_bore2pix_hwp',
    arg=(
        qp_memory_t_p,
        quat_t,
        arr,
        quat_t_p,
        quat_t_p,
        ct.c_int,
        warri,
        warr,
        warr,
        ct.c_int,
    ),
)
setargs(
    'qp_bore2pixpa',
    arg=(qp_memory_t_p, quat_t, arr, quat_t_p, ct.c_int, warri, warr, ct.c_int),
)
setargs(
    'qp_bore2pixpa_hwp',
    arg=(
        qp_memory_t_p,
        quat_t,
        arr,
        quat_t_p,
        quat_t_p,
        ct.c_int,
        warri,
        warr,
        ct.c_int,
    ),
)

setargs(
    'qp_get_interp_valn', arg=(qp_memory_t_p, ct.c_int, arr, arr, arr, warr, ct.c_int)
)

setargs(
    'qp_set_iers_bulletin_a',
    arg=(qp_memory_t_p, ct.c_int, ct.c_int, arr, arr, arr),
    res=ct.c_int,
)
setargs(
    'qp_get_iers_bulletin_a',
    arg=(
        qp_memory_t_p,
        ct.c_double,
        ct.POINTER(ct.c_double),
        ct.POINTER(ct.c_double),
        ct.POINTER(ct.c_double),
    ),
    res=ct.c_int,
)


def qp_get_bulletin_a(mem, mjd):
    dut1 = ct.c_double()
    x = ct.c_double()
    y = ct.c_double()

    libqp.qp_get_iers_bulletin_a(mem, mjd, ct.byref(dut1), ct.byref(x), ct.byref(y))
    return dut1.value, x.value, y.value


setargs('qp_refraction', arg=(ct.c_double,) * 5, res=ct.c_double)
setargs('qp_update_ref', arg=(qp_memory_t_p, quat_t), res=ct.c_double)


def get_vec_mode(map_in=None, pol=True, vpol=False):
    if pol is None:
        pol = True
    if map_in is False:
        return QP_VEC_NONE
    if map_in is None:
        return QP_VEC_VPOL if vpol else QP_VEC_POL if pol else QP_VEC_TEMP
    n = len(map_in)
    if n not in vec_modes:
        raise ValueError('Unrecognized map')
    mode = vec_modes[n]
    if isinstance(mode, dict):
        mode = mode[bool(pol)]
    return mode


def get_proj_mode(proj_in=None, pol=True, vpol=False):
    if pol is None:
        pol = True
    if proj_in is False:
        return QP_PROJ_NONE
    if proj_in is None:
        return QP_PROJ_VPOL if vpol else QP_PROJ_POL if pol else QP_PROJ_TEMP
    n = len(proj_in)
    if n not in proj_modes:
        raise ValueError('Unrecognized proj')
    mode = proj_modes[n]
    if isinstance(mode, dict):
        mode = mode[bool(pol)]
    return mode


# **********************************************************************
# Mapping functions
# **********************************************************************

# initialize detectors
setargs(
    'qp_init_det', arg=(quat_t, ct.c_double, ct.c_double, mueller_t), res=qp_det_t_p
)
setargs('qp_default_det', res=qp_det_t_p)
setargs('qp_init_det_tod', arg=(qp_det_t_p, ct.c_size_t))
setargs('qp_init_det_tod_from_array', arg=(qp_det_t_p, arr, ct.c_size_t, ct.c_int))
setargs('qp_init_det_flag', arg=(qp_det_t_p, ct.c_size_t))
setargs('qp_init_det_flag_from_array', arg=(qp_det_t_p, arrf, ct.c_size_t, ct.c_int))
setargs('qp_free_det', arg=qp_det_t_p)
setargs(
    'qp_init_detarr',
    arg=(quat_t_p, arr, arr, mueller_t_p, ct.c_size_t),
    res=qp_detarr_t_p,
)
setargs('qp_init_detarr_tod', arg=(qp_detarr_t_p, ct.c_size_t))
setargs(
    'qp_init_detarr_tod_from_array_1d', arg=(qp_detarr_t_p, arr, ct.c_size_t, ct.c_int)
)
setargs('qp_init_detarr_flag', arg=(qp_detarr_t_p, ct.c_size_t))
setargs(
    'qp_init_detarr_flag_from_array_1d', arg=(qp_detarr_t_p, arr, ct.c_size_t, ct.c_int)
)
setargs('qp_free_detarr', arg=qp_detarr_t_p)

# initialize pointing
setargs('qp_init_point', arg=(ct.c_size_t, ct.c_int, ct.c_int), res=qp_point_t_p)
setargs(
    'qp_init_point_from_arrays',
    arg=(quat_t_p, arr, quat_t_p, ct.c_size_t, ct.c_int),
    res=qp_point_t_p,
)
setargs('qp_free_point', arg=qp_point_t_p)

# initialize maps
setargs(
    'qp_init_map',
    arg=(ct.c_size_t, ct.c_size_t, qp_vec_mode, qp_proj_mode),
    res=qp_map_t_p,
)
setargs(
    'qp_init_map_from_arrays_1d',
    arg=(arr, arr, ct.c_size_t, ct.c_size_t, qp_vec_mode, qp_proj_mode, ct.c_int),
    res=qp_map_t_p,
)
setargs('qp_init_map_from_map', arg=(qp_map_t_p, ct.c_int, ct.c_int), res=qp_map_t_p)
setargs('qp_free_map', arg=qp_map_t_p)
setargs('qp_reshape_map', arg=qp_map_t_p, res=ct.c_int)
setargs('qp_init_map_pixhash', arg=(qp_map_t_p, larr, ct.c_size_t), res=ct.c_int)

# tod -> map
setargs('qp_add_map', arg=(qp_map_t_p, qp_map_t_p), res=ct.c_int)
setargs(
    'qp_tod2map1',
    arg=(qp_memory_t_p, qp_det_t_p, qp_point_t_p, qp_map_t_p),
    res=ct.c_int,
)
setargs(
    'qp_tod2map',
    arg=(qp_memory_t_p, qp_detarr_t_p, qp_point_t_p, qp_map_t_p),
    res=ct.c_int,
)

# map -> tod
setargs(
    'qp_map2tod1',
    arg=(qp_memory_t_p, qp_det_t_p, qp_point_t_p, qp_map_t_p),
    res=ct.c_int,
)
setargs(
    'qp_map2tod',
    arg=(qp_memory_t_p, qp_detarr_t_p, qp_point_t_p, qp_map_t_p),
    res=ct.c_int,
)

# **********************************************************************
# Parameters
# **********************************************************************


def check_set_float(val):
    if not np.isscalar(val):
        raise TypeError('val must be a scalar value')
    return float(val)


def check_set_int(val):
    if not np.isscalar(val):
        raise TypeError('val must be a scalar value')
    return int(val)


def check_pass(val):
    return val


def set_rfunc(state):
    f = libqp['qp_set_rate_%s' % state]
    f.argtypes = (qp_memory_t_p, ct.c_double)
    f.restype = None
    return f


def reset_rfunc(state):
    f = libqp['qp_reset_rate_%s' % state]
    f.argtypes = (qp_memory_t_p,)
    f.restype = None
    return f


def get_rfunc(state):
    f = libqp['qp_get_rate_%s' % state]
    f.argtypes = (qp_memory_t_p,)
    f.restype = ct.c_double
    return f


def check_set_state(state):
    rdict = {'always': QP_DO_ALWAYS, 'once': QP_DO_ONCE, 'never': QP_DO_NEVER}
    state = rdict.get(state, state)
    return check_set_float(state)


def check_get_state(state):
    rdict = {QP_DO_ALWAYS: 'always', QP_DO_ONCE: 'once', QP_DO_NEVER: 'never'}
    state = rdict.get(state, state)
    return state


states = ['lonlat', 'npb', 'erot', 'daber', 'aaber', 'wobble', 'dut1', 'ref']
inv_states = [k + '_inv' for k in states]
state_funcs = dict()
for s in states + inv_states:
    k = 'rate_%s' % s
    state_funcs[k] = dict()
    state_funcs[k]['set'] = set_rfunc(s)
    state_funcs[k]['reset'] = reset_rfunc(s)
    state_funcs[k]['get'] = get_rfunc(s)
    state_funcs[k]['check_set'] = check_set_state
    state_funcs[k]['check_get'] = check_get_state


def set_wfunc(par):
    f = libqp['qp_set_weather_%s' % par]
    f.argtypes = (qp_memory_t_p, ct.c_double)
    f.restype = None
    return f


def get_wfunc(par):
    f = libqp['qp_get_weather_%s' % par]
    f.argtypes = (qp_memory_t_p,)
    f.restype = ct.c_double
    return f


weather_params = ['temperature', 'pressure', 'humidity', 'frequency']
weather_funcs = dict()
for w in weather_params:
    weather_funcs[w] = dict()
    weather_funcs[w]['set'] = set_wfunc(w)
    weather_funcs[w]['get'] = get_wfunc(w)
    weather_funcs[w]['check_set'] = check_set_float
    weather_funcs[w]['check_get'] = check_pass


def set_ofunc(option):
    f = libqp['qp_set_opt_%s' % option]
    f.argtypes = (qp_memory_t_p, ct.c_int)
    f.restype = None
    return f


def get_ofunc(option):
    f = libqp['qp_get_opt_%s' % option]
    f.argtypes = (qp_memory_t_p,)
    f.restype = ct.c_int
    return f


opts = {
    'accuracy': {0: 'high', 1: 'low'},
    'polconv': {0: ['healpix', 'cosmo'], 1: 'iau'},
    'pix_order': {0: 'ring', 1: ['nest', 'nested']},
}
defaults = {'accuracy': 0, 'polconv': 0, 'pix_order': 0}


def check_set_dict(opt):
    def func(val):
        if val is None:
            return defaults[opt]
        if isinstance(val, str):
            val = val.lower()
        for k in opts[opt]:
            v = opts[opt][k]
            if isinstance(v, list):
                if val in v:
                    return k
            else:
                if val == v:
                    return k
            if val == k:
                return k
        return defaults[opt]

    return func


def check_get_dict(opt):
    def func(val):
        for k in opts[opt]:
            v = opts[opt][k]
            if val == k:
                if isinstance(v, list):
                    return v[0]
                return v
        raise ValueError('unrecognized value {} for option {}'.format(val, opt))

    return func


check_set_accuracy = check_set_dict('accuracy')
check_get_accuracy = check_get_dict('accuracy')

check_set_polconv = check_set_dict('polconv')
check_get_polconv = check_get_dict('polconv')

check_set_pix_order = check_set_dict('pix_order')
check_get_pix_order = check_get_dict('pix_order')


def check_get_bool(val):
    return bool(val)


def check_set_bool(val):
    return int(bool(val))


check_set_mean_aber = check_set_bool
check_get_mean_aber = check_get_bool

check_set_fast_math = check_set_bool
check_get_fast_math = check_get_bool

check_set_fast_pix = check_set_bool
check_get_fast_pix = check_get_bool

check_set_error_missing = check_set_bool
check_get_error_missing = check_get_bool

check_set_nan_missing = check_set_bool
check_get_nan_missing = check_get_bool

check_set_interp_missing = check_set_bool
check_get_interp_missing = check_get_bool

check_set_interp_pix = check_set_bool
check_get_interp_pix = check_get_bool


def check_set_num_threads(nt):
    if nt is None:
        return 0
    return int(nt)


def check_get_num_threads(nt):
    return nt


def check_set_thread_num(tn):
    if tn is None:
        return 0
    return int(nt)


def check_get_thread_num(tn):
    return tn


options = [
    'accuracy',
    'mean_aber',
    'fast_math',
    'polconv',
    'pix_order',
    'interp_pix',
    'fast_pix',
    'error_missing',
    'nan_missing',
    'interp_missing',
    'num_threads',
    'thread_num',
]
option_funcs = dict()
for p in options:
    option_funcs[p] = dict()
    option_funcs[p]['set'] = set_ofunc(p)
    option_funcs[p]['get'] = get_ofunc(p)
    option_funcs[p]['check_set'] = globals()['check_set_%s' % p]
    option_funcs[p]['check_get'] = globals()['check_get_%s' % p]


def set_pfunc(par):
    f = libqp['qp_set_%s' % par]
    f.argtypes = (qp_memory_t_p, ct.c_double)
    f.restype = None
    return f


def get_pfunc(par):
    f = libqp['qp_get_%s' % par]
    f.argtypes = (qp_memory_t_p,)
    f.restype = ct.c_double
    return f


double_params = ['ref_delta', 'dut1']
double_funcs = dict()
for p in double_params:
    double_funcs[p] = dict()
    double_funcs[p]['set'] = set_pfunc(p)
    double_funcs[p]['get'] = get_pfunc(p)
    double_funcs[p]['check_set'] = check_set_float
    double_funcs[p]['check_get'] = check_pass

qp_funcs = dict()
qp_funcs['rates'] = state_funcs
qp_funcs['options'] = option_funcs
qp_funcs['weather'] = weather_funcs
qp_funcs['params'] = double_funcs

# **********************************************************************
# Argument checking
# **********************************************************************


def check_flags(arg):
    return (
        arg.flags['C_CONTIGUOUS']
        | arg.flags['OWNDATA'] << 1
        | arg.flags['WRITEABLE'] << 2
        | arg.flags['ALIGNED'] << 3
    )


def check_input(
    name,
    arg,
    shape=None,
    quat=False,
    dtype=np.double,
    inplace=True,
    fill=0,
    allow_transpose=True,
    allow_tuple=True,
    output=False,
):
    """
    Ensure input argument is an aligned array of the right type and shape.

    Arguments
    ---------
    name : string
        Name of the argument
    arg : array_like
        The argument itself
    shape : tuple, optional
        If supplied, ensure `arg` has this shape.
    quat : bool, optional
        If True, ensure that the last dimension of the array has length 4.
    dtype : numpy.dtype, optional
        Ensure `arg` is of this dtype.  Default: numpy.double.
    inplace : bool, optional
        If False make sure that a copy of the input `arg` is made prior to
        returning.  Otherwise, the output `arg` may share memory with the
        input.
    fill : scalar, optional
        If the input is missing or empty, fill with this value.
        If None, leave the array empty.
    allow_transpose : bool, optional
        If True, transpose the input array if the transposed shape matches
        `shape`.
    allow_tuple : bool, optional
        If True, `numpy.vstack` the input `arg` if it is a tuple.
    output : bool, optional
        If True, ensure that the output `arg` is a writeable array.
        Otherwise, the array is only ensured to be aligned and
        C-contiguous.

    Returns
    -------
    arg :
        Aligned, contiguous and properly typed and shaped array for
        passing to the C library.
    """
    if arg is None:
        if shape is None:
            raise ValueError('need shape to initialize input!')
        if fill is None:
            arg = np.empty(shape, dtype=dtype)
        else:
            arg = fill * np.ones(shape, dtype=dtype)
    if isinstance(arg, tuple) and allow_tuple:
        arg = np.vstack(arg)
    if np.isscalar(arg):
        arg = np.array(arg)
    if not isinstance(arg, np.ndarray):
        raise TypeError('input {} must be of type numpy.ndarray'.format(name))
    if quat and arg.shape[-1] != 4:
        raise ValueError('input {} is not a valid quaternion array')
    if shape is not None:
        if arg.shape != shape:
            if arg.T.shape == shape and allow_transpose:
                arg = arg.T
            else:
                try:
                    _, arg = np.broadcast_arrays(np.empty(shape), arg)
                    # ensure writeable flag is set properly if necessary
                    # broadcast arrays will not be writable in the future
                    if inplace or output:
                        arg = arg.copy()
                    else:
                        arg.flags['WRITEABLE'] = False
                except ValueError:
                    s = 'input {} of shape {} cannot be broadcast to shape {}'
                    raise ValueError(s.format(name, arg.shape, shape))
    istat = check_flags(arg)
    arg = np.require(arg, dtype, list('AC' + 'W' * output))
    ostat = check_flags(arg)
    if istat == ostat and inplace is False:
        return arg.copy()
    return arg


def check_inputs(*args, **kwargs):
    """
    Ensure that a group of input arguments have the same shape by broadcasting.

    Arguments
    ---------
    args :
        A list of broadcastable array_like arguments.
    kwargs :
        Dictionary of arguments to pass to `check_input`.

    Returns
    -------
    args :
        A list of broadcast and properly aligned/shaped/typed arrays for
        passing to the C library as a set of timestreams.
    """
    args = [arg if arg is not None else 0 for arg in args]
    if 'shape' not in kwargs:
        kwargs['shape'] = np.broadcast(*args).shape
    return [
        check_input('argument {}'.format(ix), np.atleast_1d(x), **kwargs)
        for ix, x in enumerate(args)
    ]


def check_output(
    name,
    arg=None,
    shape=None,
    quat=False,
    dtype=np.double,
    inplace=True,
    fill=None,
    allow_transpose=True,
    allow_tuple=True,
    **kwargs
):
    """
    Ensure that the output argument is properly aligned/shaped/typed.
    Check input kwargs to see if a pointer to the output array
    has been passed in.

    Arguments
    ---------
    name : string
        Name of the argument
    arg : array_like, optional
        The argument itself.  If not supplied, kwargs is checked.
    shape : tuple, optional
        If supplied, ensure `arg` has this shape.  If `arg` is None,
        this argument is required.
    dtype : numpy.dtype, optional
        Ensure `arg` is of this dtype.  Default: numpy.double.
    inplace : bool, optional
        If False make sure that a copy of the input `arg` is made prior to
        returning.  Otherwise, the output `arg` may share memory with the
        input.
    fill : scalar, optional
        If the input is missing or empty, fill with this value.
        If None, leave the array empty.
    allow_transpose : bool, optional
        If True, transpose the input array if the transposed shape matches
        `shape`.
    allow_tuple : bool, optional
        If True, `numpy.vstack` the input `arg` if it is a tuple.
    kwargs :
        Any remaining input arguments.  If `arg` is None,
        `kwargs` is searched for the `name` key.  If not found, a
        an empty array is created of the appropriate shape.

    Returns
    -------
    arg :
        Aligned, contiguous, writeable and properly typed and shaped array
        for passing to the C library.
    """
    if arg is None:
        arg = kwargs.pop(name, None)
    return check_input(
        name, arg, shape, quat, dtype, inplace, fill, allow_transpose, allow_tuple, True
    )



#!/usr/bin/env python
#
# The following IERS data are stored:
#
#   * Bull. A PM-x (sec. of arc)
#   * Bull. A PM-y (sec. of arc)
#   * Bull. A UT1-UTC (sec. of time)
#
# 2012 Mike Nolta <mike@nolta.net>
# 2015 ASR <arahlin@princeton.edu>
# 2017 ASR <arahlin@fnal.gov> -- use astropy utilities for auto update

header = """
/*
 IERS Bulletin A interpolation
 2012 Mike Nolta <mike@nolta.net>
 2015 ASR <arahlin@princeton.edu>
 2017 ASR <arahlin@fnal.gov>
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qpoint.h"

"""

footer = r"""
int qp_get_iers_bulletin_a( qp_memory_t *mem, double mjd,
                           double *dut1, double *x, double *y )
{
  qp_bulletina_t *B = &mem->bulletinA;

  if (B->entries == NULL) {
    B->entries = bulletinA_factory;
    B->mjd_min = mjd_min_factory;
    B->mjd_max = mjd_max_factory;
  }
  if ( (B->mjd_min <= mjd) && (mjd < B->mjd_max) )
  {
    double mjd_floor;
    double r = modf( mjd, &mjd_floor );
    int k = (int) mjd_floor - B->mjd_min;
    qp_bulletina_entry_t a = B->entries[k];
    qp_bulletina_entry_t b = B->entries[k+1];
    // Detect leap seconds
    double leap = b.dut1 - a.dut1;
    if (leap > 0.5)
      leap = 1.;
    else if (leap < -0.5)
      leap = -1.;
    else
      leap = 0;

    *dut1 = (1-r)*a.dut1 + r*(b.dut1-leap);
    *x = (1-r)*a.x + r*b.x;
    *y = (1-r)*a.y + r*b.y;
  }
  else
  {
    // silent failure!
    *dut1 = 0;
    *x = 0;
    *y = 0;
    return 1;
  }

  return 0;
}

int qp_set_iers_bulletin_a( qp_memory_t *mem, int mjd_min_, int mjd_max_,
                           double *dut1, double *x, double *y)
{
  qp_bulletina_t *B = &mem->bulletinA;

  if (dut1 == NULL) {
    // Destroy
    if (B->entries != bulletinA_factory && B->entries != NULL) {
      free(B->entries);
      B->entries = NULL;
    }
    return 0;
  }
  // Create
  B->mjd_min = mjd_min_;
  B->mjd_max = mjd_max_;

  int n_mjd = B->mjd_max - B->mjd_min + 1;
  B->entries = malloc(n_mjd*sizeof(*(B->entries)));
  if (B->entries == NULL)
    return 1;

  for (int k=0; k<n_mjd; k++) {
    B->entries[k].x = x[k];
    B->entries[k].y = y[k];
    B->entries[k].dut1 = dut1[k];
  }
  return 0;
}

int qp_copy_iers_bulletin_a(qp_memory_t *memdest, qp_memory_t *memsrc) {
  qp_bulletina_t *bdest = &memdest->bulletinA;
  qp_bulletina_t *bsrc = &memsrc->bulletinA;


  // nothing to do, since both source and dest point to factory
  if (bdest->entries == bsrc->entries && bsrc->entries == bulletinA_factory)
    return 0;

  if (bsrc->entries != bulletinA_factory && bsrc->entries != NULL) {
    int n_mjd = bsrc->mjd_max - bsrc->mjd_min + 1;
    bdest->mjd_max = bsrc->mjd_max;
    bdest->mjd_min = bsrc->mjd_min;
    if (bdest->entries != bulletinA_factory && bdest->entries != NULL) {
      if (bdest->entries != bsrc->entries)
        free(bdest->entries);
      bdest->entries = NULL;
    }
    bdest->entries = malloc(n_mjd*sizeof(*(bdest->entries)));
    if (bdest->entries == NULL)
      return 1;
    memcpy(bdest->entries, bsrc->entries, n_mjd*sizeof(*(bsrc->entries)));
  }

  return 0;
}
"""

if __name__ == '__main__':

    f = open( "qp_iers_bulletin_a.c", "w" )
    f.write( header )

    f.write( "static qp_bulletina_entry_t bulletinA_factory[] = {\n" )

    try:
        from astropy.utils.iers import IERS_Auto

    except ImportError:
        from warnings import warn, simplefilter
        simplefilter('always')
        warn('Compatible Astropy not found, creating an empty IERS-A table. '
             'Install astropy v1.2 or newer for accurate polar motion '
             'and UT1 corrections',
             ImportWarning)

        f.write( " {0., 0., 0.}, // NULL\n")
        mjd_min = 0
        mjd_max = 0

    else:
        import numpy as np

        columns = ['year', 'month', 'day', 'MJD', 'PM_x', 'PM_y', 'UT1_UTC']
        iers_table = IERS_Auto.open()[columns].as_array()

        # check year
        year = iers_table['year'] + 1900
        wraps, = np.where(np.ediff1d(year) < 0)
        for idx in wraps:
            year[idx + 1:] += 100
        iers_table['year'] = year
        iers_table = iers_table[year >= 2000]

        # check MJD
        mjds = iers_table['MJD']
        mjd_min = int(mjds.min())
        mjd_max = int(mjds.max())

        # populate factory table
        for row in iers_table:
            f.write( "  {%.6f, %.6f, %.7f},  // %d-%02d-%02d\n" % (
                row['PM_x'], row['PM_y'], row['UT1_UTC'],
                row['year'], row['month'], row['day']) )

    f.write( "};\n\n" )
    
    f.write( "static const int mjd_min_factory = %d;\n" % mjd_min )
    f.write( "static const int mjd_max_factory = %d;\n" % mjd_max )
    
    f.write( footer )
    f.close()



"""A lightweight quaternion-based library for efficient telescope pointing.
"""

from ._version import __version__


def version():
    """Print version string"""
    return __version__


from . import tools
from .qpoint_class import *
from .qmap_class import *



import qpoint as qp
import numpy as np
import healpy as hp
from matplotlib import pyplot as plt

# initialize, maybe change a few options from their defaults
Q = qp.QMap(
    update_iers=True,
    nside=256,
    pol=True,
    accuracy="low",
    fast_math=True,
    mean_aber=True,
    num_threads=4,
)

# dumb simulation
n = 100000
ctime = 1418662800.0 + np.arange(n) / 100.0
az = 100.0 + 40.0 * np.sin(2 * np.pi * np.arange(n) / 4000.0)
el = 32.0 + 10.0 * np.mod(np.arange(n, dtype=float), 500000.0) / 500000.0
el = np.floor(el / 0.1) * 0.1
pitch = None
roll = None
lat = -77.6 * np.ones_like(ctime)
lon = 165.7 - np.arange(n) * 3 / 850000.0

# step waveplate twice a day...
lmst = Q.lmst(ctime, lon)
hwp = np.ones_like(lmst)
hwp[lmst <= 12] = 22.5
hwp[lmst > 12] = 45.0
q_hwp = Q.hwp_quat(hwp)

# calculate boresight quaternions
q_bore = Q.azel2bore(az, el, pitch, roll, lon, lat, ctime)

# store pointing for mapmaking
Q.init_point(q_bore=q_bore, q_hwp=q_hwp)

cls = np.loadtxt("wmap7_r0p03_lensed_uK_ext.txt", unpack=True)
ell, cls = cls[0], cls[1:]
map_in = np.vstack(hp.synfast(cls, nside=512, pol=True, new=True))
print(map_in.mean(), map_in.sum())

# initialize source map
Q.init_source(map_in, pol=True)

# detector offsets
d_psi0 = [0, 45.2, 90.5, 135]
n = len(d_psi0)
d_az = [0] * n + [1] * n + [2] * n + [-1] * n + [-2] * n
d_el = [0] * n + [-1] * n + [-2] * n + [1] * n + [2] * n
d_psi = d_psi0 * 5
d_az, d_el = (d_az + d_az[n:], d_el + d_az[n:])
d_psi += d_psi[n:]

q_off_list = Q.det_offset(d_az, d_el, d_psi)

# run
print("generate tod")
tod = Q.to_tod(q_off_list)
print(tod.mean(), tod.sum())

# initialize and calculate hits and data maps
print("bin to map")
vec, proj = Q.from_tod(q_off_list, tod=tod)
print("map stats")
print(vec.mean(), vec.sum())
print(proj.mean(), proj.sum())

print("add to map")

# add noise
tod2 = 10 * np.random.randn(*tod.shape)

# update map, but don't double-count the hits map
Q.from_tod(q_off_list, tod=tod2, count_hits=False)
print("map stats")
print(vec.mean(), vec.sum())
print(proj.mean(), proj.sum())

print("solving")

# solve
cond = Q.proj_cond()
print(cond.min(), cond.max())
map_out = Q.solve_map(fill=np.nan)
print(np.nanmin(map_out), np.nanmax(map_out))

print("plotting")

# plot stuff
opt = dict(lonra=[-170, -70], latra=[-50, -20], coord="C", cbar="h")
lopt = dict(min=-300, max=300, **opt)

hp.cartview(map_in[0], title="input map", unit="Temperature [uK]", **lopt)
hp.cartview(proj[0], title="hits", **opt)
hp.cartview(cond, title="condition number", **opt)
hp.cartview(map_out[0], title="solved map", unit="Temperature[uK]", **lopt)

# difference map
md = map_out[0] - hp.ud_grade(map_in[0], 256)
md[map_out[0] == 0] = np.inf
hp.cartview(md, title="difference", unit="Temperature [uK]", min=-20, max=20, **opt)

plt.show()



import qpoint as qp
import numpy as np

# initialize, maybe change a few options from their defaults
Q = qp.QPoint(update_iers=True, accuracy="low", fast_math=True, mean_aber=True)

print("simulate pointing")

# dumb simulation
n = 100000
ctime = 1418662800.0 + np.arange(n) / 100.0
az = 100.0 + 40.0 * np.sin(2 * np.pi * np.arange(n) / 4000.0)
el = 32.0 + 10.0 * np.mod(np.arange(n, dtype=float), 500000.0) / 500000.0
el = np.floor(el / 0.1) * 0.1
pitch = None  # np.zeros_like(ctime)
roll = None  # np.zeros_like(ctime)
lat = -77.6 * np.ones_like(ctime)
lon = 165.7 - np.arange(n) * 3 / 850000.0

# step waveplate twice a day...
lmst = Q.lmst(ctime, lon)
hwp = np.ones_like(lmst)
hwp[lmst <= 12] = 22.5
hwp[lmst > 12] = 45.0
q_hwp = Q.hwp_quat(hwp)

# calculate boresight quaternions
q_bore = Q.azel2bore(az, el, pitch, roll, lon, lat, ctime)

# detector offset in degrees
delta_az = 1.0
delta_el = -1.0
delta_psi = 22.5

q_off = Q.det_offset(delta_az, delta_el, delta_psi)

# calculate detector pointing
ra, dec, sin2psi, cos2psi = Q.bore2radec(q_off, ctime, q_bore, q_hwp=q_hwp)

# calculate healpix pixel numbers
pix, sin2psi, cos2psi = Q.bore2pix(q_off, ctime, q_bore, q_hwp=q_hwp)

print(ra.min(), ra.max())
print(dec.min(), dec.max())
print(pix.min(), pix.max())

# Check round-trip conversions
ra1, dec1, pa1 = Q.bore2radec(np.asarray([1, 0, 0, 0]), ctime, q_bore, return_pa=True)
az2, el2, pa2 = Q.radec2azel(ra1, dec1, pa1, lon, lat, ctime)

az_diff = az2 - az
el_diff = el2 - el
pa_diff = pa2

print(az_diff.min(), az_diff.max())
print(el_diff.min(), el_diff.max())
print(pa_diff.min(), pa_diff.max())



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from warnings import warn
from . import _libqpoint as lib
from ._libqpoint import libqp as qp
from ._libqpoint import check_input, check_inputs, check_output

__all__ = ['QPoint', 'check_input', 'check_inputs', 'check_output']


class QPoint(object):
    def __init__(self, update_iers=False, **kwargs):
        """
        Initialize a `QPoint` memory instance for keeping track of pointing
        corrections over time.

        If `update_iers` is `True`, attempt to call the method
        :meth:`update_bulletin_a` to update the internal IERS-A database
        (requires `astropy` version 1.2 or newer).

        Any remaining keyword arguments are passed to :meth:`set` to update
        memory.
        """

        # initialize memory
        self._memory = qp.qp_init_memory()

        # try to update IERS-A bulletin
        if update_iers:
            self.update_bulletin_a()

        # collect all parameter functions
        self._funcs = lib.qp_funcs
        self._all_funcs = dict()
        for k in self._funcs:
            self._all_funcs.update(**self._funcs[k])

        # set any requested parameters
        self.set(**kwargs)

    def print_memory(self):
        """
        Print current memory state in C.
        """
        qp.qp_print_memory(self._memory)

    def __del__(self):
        """
        Free memory before deleting the object
        """
        qp.qp_free_memory(self._memory)

    def _set(self, key, val):
        """
        Set a single parameter to the given value.  See :meth:`set` for a list
        of parameter names.
        """
        if key not in self._all_funcs:
            raise KeyError('Unknown parameter {}'.format(key))
        val = self._all_funcs[key]['check_set'](val)
        self._all_funcs[key]['set'](self._memory, val)

    def _get(self, key):
        """
        Get the value for a single parameter.  See :meth:`set` for a list of
        parameter names.
        """
        if key not in self._all_funcs:
            raise KeyError('Unknown parameter {}'.format(key))
        val = self._all_funcs[key]['get'](self._memory)
        return self._all_funcs[key]['check_get'](val)

    def set(self, **kwargs):
        """
        Set computation options.

        Arguments
        ---------
        rate_daber : {'never', 'once', 'always'}, or float
            Rate at which the diurnal aberration correction is applied in seconds
            (NB: this can only be applied always or never)
        rate_lonlat : {'never', 'once', 'always'}, or float
            Rate at which observer's lon and lat are updated
        rate_wobble : {'never', 'once', 'always'}, or float
            Rate at which the polar motion correction is updated
            (NB: these are not estimated for dates beyond a year from now)
        rate_dut1 : {'never', 'once', 'always'}, or float
            Rate at which the ut1-utc correction is updated
            (NB: this is not estimated for dates beyond a year from now)
        rate_erot : {'never', 'once', 'always'}, or float
            Rate at which the earth's rotation angle is updated
        rate_npb : {'never', 'once', 'always'}, or float
            Rate at which the nutation/precession/frame-bias terms are updated
        rate_aaber : {'never', 'once', 'always'}, or float
            Rate at which the annual aberration correction
            (due to the earth's orbital velocity) is updated
        rate_ref : {'never', 'once', 'always'}, or float
            Rate at which the refaction correction is updated (NB: this
            correction can also be updated manually -- see :meth:`refraction`)
        accuracy : 'low' or 'high'
            If 'low', use a truncated form (2000b) for the NPB correction,
            which is much faster but less accurate. If 'high' (default), use
            the full 2006/2000a form.
        mean_aber : bool
            If True, apply the aberration correction as an average for the
            entire field of view.  This is gives a 1-2 arcsec deviation
            at the edges of the SPIDER field of view.
        fast_math : bool
            If True, use polynomial approximations for trig functions
        polconv : 'cosmo' or 'iau'
            Specify the 'cosmo' or 'iau' polarization convention
        pix_order : 'nest' or 'ring'
            HEALPix pixel ordering
        interp_pix : bool
            If True, interpolate between pixels in scanning the source map.
        fast_pix : bool
            If True, use `vec2pix` to get pixel number directly from the
            quaternion instead of `ang2pix` from ra/dec.
        error_missing : bool
            If True, raise an error if reading/writing missing pixels.
        nan_missing : bool
            If True, fill samples from missing pixels with NaN.
            Only used if `error_missing` is False.
        interp_missing : bool
            If True and `interp_pix` is True, drop missing neighbors
            and reweight remaining neighbors.  Overrides `nan_missing`.
        num_threads : bool
             Number of openMP threads to use for mapmaking.
        temperature : float
            Ambient temperature, Celcius. For computing refraction corrections.
        pressure : float
            Ambient pressure, mbar. For computing refraction corrections.
        humidity : float
            Relative humidity, fraction. For computing refraction corrections.
        frequency : float
            Observer frequency, GHz. For computing refraction corrections.
        dut1 : float
            UT1 correction
        ref_delta : float
            Refraction correction
        """

        for k in kwargs:
            try:
                self._set(k, kwargs[k])
            except KeyError:
                continue

    def get(self, *args):
        """
        Returns a dictionary of the requested state parameters.  If no
        parameters are supplied, then all are returned.  If a single parameter
        is supplied, then just that value is returned.  See :meth:`set` for a
        list of parameter names.

        Can also select 'options', 'rates', 'weather', or 'params' to return
        all of that subset of parameters.
        """
        state = dict()
        if not len(args):
            for k in self._funcs:
                state[k] = dict()
                for kk in self._funcs[k]:
                    state[k][kk] = self._get(kk)
            return state

        for arg in args:
            if arg in self._funcs:
                state[arg] = dict()
                for k in self._funcs[arg]:
                    state[arg][k] = self._get(k)
            else:
                state[arg] = self._get(arg)
        if len(args) == 1:
            return state[args[0]]
        return state

    def reset_rates(self):
        """
        Reset update counters for each state.  Useful to force an updated
        correction term at the beginning of each chunk.
        """
        qp.qp_reset_rates(self._memory)

    def reset_inv_rates(self):
        """
        Reset update counters for each inverse state.  Useful to force an updated
        correction term at the beginning of each chunk.
        """
        qp.qp_reset_inv_rates(self._memory)

    def refraction(self, *args, **kwargs):
        """refraction(q, **kwargs)
        refraction(delta)

        Update refraction parameters

        Arguments
        ---------
        q : quaternion or array of quaternions
            Observer orientation in horizon coordinates
        temperature : float
            Ambient temperature, Celcius
        pressure : float
            Ambient pressure, mbar
        humidity : float
            Ambient relative humidity, fraction
        frequency : float
            Observing frequency, GHz
        delta : float
            The refraction correction itself, in degrees

        Returns
        -------
        delta : array_like
            Refraction correction computed at each input orientation

        Notes
        -----
        If `q` is given, then the refraction correction in degrees
        is calculated, stored and returned after updating any other given
        parameters. Otherwise, the correction is returned w/out recalculating.

        Alternatively, if a single numerical argument, or the `delta` keyword
        argument is given, then the correction is stored with this value
        instead of being recalculated.

        Numpy-vectorized for the `q` argument.  Note that this is not
        an efficient vectorization, and only the last calculated value is
        stored for use in the coordinate conversion functions.
        """

        if len(args) == 1 and len(kwargs) == 0:
            v = args[0]
            if np.isscalar(v):
                self._set('ref_delta', v)
                return v

        if 'delta' in kwargs:
            v = kwargs.get('delta')
            self._set('ref_delta', v)
            return v

        arg_names = ['q'] + list(self._funcs['weather'])
        for idx, a in enumerate(args):
            kwargs[arg_names[idx]] = a

        for w in self._funcs['weather']:
            if w in kwargs:
                self._set(w, kwargs.get(w))

        q = kwargs.get('q', None)
        if q is not None:

            def func(x0, x1, x2, x3):
                q = np.ascontiguousarray([x0, x1, x2, x3])
                return qp.qp_update_ref(self._memory, q)

            fvec = np.vectorize(func, [np.double])
            if q.size // 4 > 1:
                q = q.transpose()
            delta = fvec(q[0], q[1], q[2], q[3])
            if delta.shape == ():
                return delta[()]
            return delta
        return self._get('ref_delta')

    def gmst(self, ctime, **kwargs):
        """
        Return Greenwich mean sidereal time for given ctimes.
        Vectorized.

        Arguments
        ---------
        ctime : array_like
            Unix time in seconds UTC

        Returns
        -------
        gmst : array_like
            Greenwich mean sidereal time of the observer

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        ctime = check_input('ctime', np.atleast_1d(ctime))
        n = ctime.size

        gmst = check_output('gmst', shape=ctime.shape)
        qp.qp_gmstn(self._memory, ctime, gmst, n)

        if n == 1:
            return gmst[0]
        return gmst

    def lmst(self, ctime, lon, **kwargs):
        """
        Return local mean sidereal time for given ctimes and longitudes.
        Vectorized, input arguments must be broadcastable to the same shape.

        Arguments
        ---------
        ctime : array_like
            Unix time in seconds UTC
        lon : array_like
            Observer longitude (degrees)

        Returns
        -------
        lmst : array_like
            Local mean sidereal time of the observer

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        ctime, lon = check_inputs(ctime, lon)
        n = ctime.size

        lmst = check_output('lmst', shape=ctime.shape)
        qp.qp_lmstn(self._memory, ctime, lon, lmst, n)

        if n == 1:
            return lmst[0]
        return lmst

    def dipole(self, ctime, ra, dec, **kwargs):
        """
        Return dipole amplitude in the given equatorial direction.
        Vectorized, input arguments must be broadcastable to the same shape.

        Arguments
        ---------
        ctime : array_like
            Unix time in seconds UTC
        ra : array_like
            Right ascension on the sky, in degrees.
        dec : array_like
            Declination on the sky, in degrees

        Returns
        -------
        dipole : array_like
            Dipole amplitude in K

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        ctime, ra, dec = check_inputs(ctime, ra, dec)
        n = ctime.size

        dipole = check_output('dipole', shape=ctime.shape, **kwargs)
        qp.qp_dipolen(self._memory, ctime, ra, dec, dipole, n)

        if n == 1:
            return dipole[0]
        return dipole

    def bore2dipole(self, q_off, ctime, q_bore, **kwargs):
        """
        Calculate dipole timestream for given offset and boresight pointing.

        Arguments
        ---------
        q_off : quaternion
            Detector offset quaternion for a single detector, calculated using
            :meth:`det_offset`
        ctime : array_like
            Array of unix times in seconds UTC
        q_bore : quaternion or array of quaternions
            Array of quaternions encoding the boresight orientation on the sky
            (as output by :meth:`azel2radec` or similar).  Broadcastable to the
            same length as `ctime`.

        Returns
        -------
        dipole : array_like
            Dipole amplitude in K

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        q_off = check_input('q_off', q_off, shape=(4,), quat=True)
        ctime = check_input('ctime', ctime)
        n = ctime.size
        q_bore = check_input('q_bore', q_bore, shape=(n, 4), quat=True)

        dipole = check_output('dipole', shape=ctime.shape, **kwargs)
        qp.qp_bore2dipole(self._memory, q_off, ctime, q_bore, dipole, n)

        if n == 1:
            return dipole[0]
        return dipole

    def det_offset(self, delta_az, delta_el, delta_psi):
        """
        Return quaternion corresponding to the requested detector centroid
        offset from boresight.  Vectorized, input arguments must be
        broadcastable to the same shape.

        Arguments
        ---------
        delta_az : array_like
            Azimuthal centroid offset of the detector in degrees
        delta_el : array_like
            Elevation centroid offset of the detector in degrees
        delta_psi : array_like
            Polarization offset of the detector from vertical in degrees

        Returns
        -------
        q : array_like
            Detector centroid offset quaternion for each detector
        """

        delta_az, delta_el, delta_psi = check_inputs(delta_az, delta_el, delta_psi)
        ndet = delta_az.size

        quat = check_output('quat', shape=(ndet, 4))
        qp.qp_det_offsetn(delta_az, delta_el, delta_psi, quat, ndet)

        if ndet == 1:
            return quat[0]
        return quat

    def bore_offset(
        self, q_bore, ang1=None, ang2=None, ang3=None, post=False, inplace=False
    ):
        """
        Apply a fixed or variable offset to the boresight quaternion.
        Input arguments must be broadcastable to the same shape.

        Arguments
        ---------
        q_bore : array_like
            boresight pointing quaternion
        ang1 : array_like, optional
            Azimuthal or ra offset in degrees
        ang2 : array_like, optional
            Elevation or dec offset in degrees
        ang3 : array_like, optional
            Position angle offset in degrees
        post : bool, optional
            If False, apply offset as an az/el/pa pre-rotation
            If True, apply offset as an ra/dec/pa post-rotation
        inplace : bool, optional
            If True, apply the rotation in-place in memory.

        Returns
        -------
        q_bore : array_like
            Offset boresight quaternion
        """
        q_bore = check_input(
            'q_bore', np.atleast_2d(q_bore), quat=True, inplace=inplace, output=True
        )
        n = q_bore.shape[0]
        if all([a is None for a in [ang1, ang2, ang3]]):
            raise ValueError('One of ang1, ang2, ang3 is required')
        ang1, ang2, ang3 = check_inputs(ang1, ang2, ang3, shape=(n,))
        qp.qp_bore_offset(self._memory, q_bore, ang1, ang2, ang3, n, int(post))
        return q_bore

    def hwp_quat(self, theta):
        """
        Return quaternion corresponding to rotation by 2 * theta, where theta is
        the physical waveplate angle. Vectorized.

        Arguments
        ---------
        theta : array_like
            HWP physical angle in degrees

        Returns
        -------
        q : array_like
            Quaternion for each hwp angle
        """
        theta = check_input('theta', np.atleast_1d(theta))
        n = theta.size

        quat = check_output('quat', shape=(n, 4))
        qp.qp_hwp_quatn(theta, quat, n)

        if n == 1:
            return quat[0]
        return quat

    def azel2bore(self, az, el, pitch, roll, lon, lat, ctime, q=None, **kwargs):
        """
        Estimate the quaternion for the boresight orientation on the sky given
        the attitude (az/el/pitch/roll), location on the earth (lon/lat) and
        ctime. Input vectors must be numpy-array-like and broadcastable to the
        same shape.

        Arguments
        ---------
        az : array_like
            Boresight azimuth in degrees
        el : array_like
            Boresight elevation in degrees
        pitch : array_like
            Boresight pitch in degrees.  If `None`, this term is ignored.
        roll : array_like
            Boresight roll in degrees.  If `None`, this term is ignored.
        lon : array_like
            Observer longitude in degrees
        lat : array_like
            Observer latitude in degrees
        ctime : array_like
            Unix time in seconds UTC
        q : array_like, optional
            Output quaternion array initialized by user.  Supply this
            for in-place computation.

        Returns
        -------
        q : array_like
            Nx4 numpy array of quaternions for each supplied timestamp.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        az, el, pitch, roll, lon, lat, ctime = check_inputs(
            az, el, pitch, roll, lon, lat, ctime
        )
        n = az.size

        # identity quaternion
        q = check_output('q', q, shape=(n, 4), fill=[1, 0, 0, 0])

        qp.qp_azel2bore(self._memory, az, el, pitch, roll, lon, lat, ctime, q, n)

        return q

    def azelpsi2bore(self, az, el, psi, pitch, roll, lon, lat, ctime, q=None, **kwargs):
        """
        Estimate the quaternion for the boresight orientation on the sky given
        the attitude (az/el/psi/pitch/roll), location on the earth (lon/lat) and
        ctime. Input vectors must be numpy-array-like and broadcastable to the
        same shape.

        This is an augmented version of azel2bore to accept rotations of the focal
        plane about the optical axis.

        Arguments
        ---------
        az : array_like
            Boresight azimuth in degrees
        el : array_like
            Boresight elevation in degrees
        psi : array_like
            Boresight rotation angle in degrees
        pitch : array_like
            Boresight pitch in degrees.  If `None`, this term is ignored.
        roll : array_like
            Boresight roll in degrees.  If `None`, this term is ignored.
        lon : array_like
            Observer longitude in degrees
        lat : array_like
            Observer latitude in degrees
        ctime : array_like
            Unix time in seconds UTC
        q : array_like, optional
            Output quaternion array initialized by user.  Supply this
            for in-place computation.

        Returns
        -------
        q : array_like
            Nx4 numpy array of quaternions for each supplied timestamp.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        az, el, psi, pitch, roll, lon, lat, ctime = check_inputs(
            az, el, psi, pitch, roll, lon, lat, ctime
        )
        n = az.size

        # identity quaternion
        q = check_output('q', q, shape=(n, 4), fill=[1, 0, 0, 0])

        qp.qp_azelpsi2bore(
            self._memory, az, el, psi, pitch, roll, lon, lat, ctime, q, n
        )

        return q

    def bore2radec(
        self,
        q_off,
        ctime,
        q_bore,
        q_hwp=None,
        sindec=False,
        return_pa=False,
        ra=None,
        dec=None,
        pa=None,
        sin2psi=None,
        cos2psi=None,
        **kwargs
    ):
        """
        Calculate the orientation on the sky for a detector offset from the
        boresight.  Detector offsets are defined assuming the boresight is
        pointed toward the horizon, and that the boresight polarization axis is
        along the vertical.

        Arguments
        ---------
        q_off : quaternion
            Detector offset quaternion for a single detector, calculated using
            :meth:`det_offset`.
        ctime : array_like
            Unix time in seconds UTC, broadcastable to shape (N,),
            the long dimension of `q_bore`.
        q_bore : quaternion or array of quaternions
            Nx4 array of quaternions encoding the boresight orientation on the
            sky (as output by :meth:`azel2radec` or equivalent)
        q_hwp : quaternion or array of quaternions, optional
            HWP angle quaternions calculated using :meth:`hwp_quat`.  Must be
            broadcastable to the same shape as `q_bore`.
        sindec : bool, optional
            If `True`, return sin(dec) instead of dec in degrees
            (default False).
        return_pa : bool, optional
            If `True`, return pa instead of sin2psi / cos2psi

        Returns
        -------
        ra : array_like
            Detector right ascension in degrees
        dec/sindec : array_like
            Detector declination in degrees or sin(dec) if `sindec` is `True`.
        pa/sin2psi : array_like
            Detector polarization orientation if `return_pa` is `True`, or
            sin(2*pa) if `return_pa` is `False`.
        cos2psi : array_like
            detector polarization orientation cos(2*pa), if `return_pa` is `False`.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.

        Pre-allocated output arguments can also be supplied as input keywords
        for in-place operation.
        """

        self.set(**kwargs)

        q_off = check_input('q_off', q_off, quat=True)
        q_bore = check_input('q_bore', np.atleast_2d(q_bore), quat=True)
        shape = (q_bore.size // 4,)
        if ctime is None:
            if not self.get('mean_aber'):
                raise ValueError('ctime required if mean_aber is False')
            ctime = np.zeros(shape, dtype=q_bore.dtype)
        ctime = check_input('ctime', ctime, shape=shape)
        pars = dict(shape=ctime.shape, dtype=np.double)
        ra = check_output('ra', ra, **pars)
        dec = check_output('dec', dec, **pars)
        if return_pa:
            if sindec:
                raise ValueError('Cannot use sindec with return_pa=True')
            pa = check_output('pa', pa, **pars)
        else:
            sin2psi = check_output('sin2psi', sin2psi, **pars)
            cos2psi = check_output('cos2psi', cos2psi, **pars)
        n = ctime.size

        if q_hwp is None:
            if return_pa:
                qp.qp_bore2radecpa(self._memory, q_off, ctime, q_bore, ra, dec, pa, n)
            elif sindec:
                qp.qp_bore2rasindec(
                    self._memory, q_off, ctime, q_bore, ra, dec, sin2psi, cos2psi, n
                )
            else:
                qp.qp_bore2radec(
                    self._memory, q_off, ctime, q_bore, ra, dec, sin2psi, cos2psi, n
                )
        else:
            q_hwp = check_input('q_hwp', q_hwp, shape=q_bore.shape)
            if return_pa:
                qp.qp_bore2radecpa_hwp(
                    self._memory, q_off, ctime, q_bore, q_hwp, ra, dec, pa, n
                )
            elif sindec:
                qp.qp_bore2rasindec_hwp(
                    self._memory,
                    q_off,
                    ctime,
                    q_bore,
                    q_hwp,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
            else:
                qp.qp_bore2radec_hwp(
                    self._memory,
                    q_off,
                    ctime,
                    q_bore,
                    q_hwp,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )

        if return_pa:
            if n == 1:
                return ra[0], dec[0], pa[0]
            return ra, dec, pa

        if n == 1:
            return ra[0], dec[0], sin2psi[0], cos2psi[0]
        return ra, dec, sin2psi, cos2psi

    def azel2radec(
        self,
        delta_az,
        delta_el,
        delta_psi,
        az,
        el,
        pitch,
        roll,
        lon,
        lat,
        ctime,
        hwp=None,
        sindec=False,
        return_pa=False,
        ra=None,
        dec=None,
        pa=None,
        sin2psi=None,
        cos2psi=None,
        **kwargs
    ):
        """
        Estimate the orientation on the sky for a detector offset from
        boresight, given the boresight attitude (az/el/pitch/roll), location on
        the earth (lon/lat) and UTC time.  Input vectors must be
        numpy-array-like and broadcastable to the same shape. Detector offsets
        are defined assuming the boresight is pointed toward the horizon, and
        that the boresight polarization axis is along the horizontal.

        Arguments
        ---------
        delta_az : float
            Azimuthal offset of the detector in degrees
        delta_el : float
            Elevation offset of the detector in degrees
        delta_psi : float
            Polarization offset of the detector in degrees
        az : array_like
            Boresight azimuth in degrees
        el : array_like
            Boresight elevation in degrees
        pitch : array_like
            Boresight pitch in degrees.  If None, this term is ignored.
        roll : array_like
            Boresight roll in degrees.  If None, this term is ignored.
        lon : array_like
            Observer longitude in degrees.
        lat : array_like
            Observer latitude in degrees.
        ctime : array_like
            Unix time in seconds UTC
        hwp : array_like, optional
            HWP angles in degrees
        sindec : bool, optional
            If `True`, return sin(dec) instead of dec in degrees (default False)
        return_pa : bool, optional
            If `True`, return pa instead of sin2psi/cos2psi

        Returns
        -------
        ra : array_like
            Detector right ascension in degrees
        dec/sindec : array_like
            Detector declination in degrees
        pa : array_like
            Detector position angle, if `return_pa` is True
        sin2psi : array_like
            Detector polarization orientation, if `return_pa` is False
        cos2psi : array_like
            Detector polarization orientation, if `return_pa` is False

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        az, el, pitch, roll, lon, lat, ctime = check_inputs(
            az, el, pitch, roll, lon, lat, ctime
        )

        ra = check_output('ra', ra, shape=az.shape, dtype=np.double)
        dec = check_output('dec', dec, shape=az.shape, dtype=np.double)
        if return_pa:
            pa = check_output('pa', pa, shape=az.shape, dtype=np.double)
        else:
            sin2psi = check_output('sin2psi', sin2psi, shape=az.shape, dtype=np.double)
            cos2psi = check_output('cos2psi', cos2psi, shape=az.shape, dtype=np.double)
        n = az.size

        if hwp is None:
            if return_pa:
                qp.qp_azel2radecpa(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    ra,
                    dec,
                    pa,
                    n,
                )
            elif sindec:
                qp.qp_azel2rasindec(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
            else:
                qp.qp_azel2radec(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
        else:
            hwp = check_input('hwp', hwp, shape=az.shape)

            if return_pa:
                qp.qp_azel2radec_hwp(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    hwp,
                    ra,
                    dec,
                    pa,
                    n,
                )
            elif sindec:
                qp.qp_azel2rasindec_hwp(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    hwp,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
            else:
                qp.qp_azel2radec_hwp(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    hwp,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )

        if return_pa:
            return ra, dec, pa
        return ra, dec, sin2psi, cos2psi

    def azelpsi2radec(
        self,
        delta_az,
        delta_el,
        delta_psi,
        az,
        el,
        psi,
        pitch,
        roll,
        lon,
        lat,
        ctime,
        hwp=None,
        sindec=False,
        return_pa=False,
        ra=None,
        dec=None,
        pa=None,
        sin2psi=None,
        cos2psi=None,
        **kwargs
    ):
        """
        Estimate the orientation on the sky for a detector offset from
        boresight, given the boresight attitude (az/el/pitch/roll), location on
        the earth (lon/lat) and UTC time.  Input vectors must be
        numpy-array-like and broadcastable to the same shape. Detector offsets
        are defined assuming the boresight is pointed toward the horizon, and
        that the boresight polarization axis is along the horizontal.

        This is agumented from azel2radec(), to accept rotations of the focal
        plane about the optical axis.

        Arguments
        ---------
        delta_az : float
            Azimuthal offset of the detector in degrees
        delta_el : float
            Elevation offset of the detector in degrees
        delta_psi : float
            Polarization offset of the detector in degrees
        az : array_like
            Boresight azimuth in degrees
        el : array_like
            Boresight elevation in degrees
        psi : array_like
            Boresight rotation in degrees
        pitch : array_like
            Boresight pitch in degrees.  If None, this term is ignored.
        roll : array_like
            Boresight roll in degrees.  If None, this term is ignored.
        lon : array_like
            Observer longitude in degrees.
        lat : array_like
            Observer latitude in degrees.
        ctime : array_like
            Unix time in seconds UTC
        hwp : array_like, optional
            HWP angles in degrees
        sindec : bool, optional
            If `True`, return sin(dec) instead of dec in degrees (default False)
        return_pa : bool, optional
            If `True`, return pa instead of sin2psi/cos2psi

        Returns
        -------
        ra : array_like
            Detector right ascension in degrees
        dec/sindec : array_like
            Detector declination in degrees
        pa : array_like
            Detector position angle, if `return_pa` is True
        sin2psi : array_like
            Detector polarization orientation, if `return_pa` is False
        cos2psi : array_like
            Detector polarization orientation, if `return_pa` is False

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        az, el, psi, pitch, roll, lon, lat, ctime = check_inputs(
            az, el, psi, pitch, roll, lon, lat, ctime
        )

        ra = check_output('ra', ra, shape=az.shape, dtype=np.double)
        dec = check_output('dec', dec, shape=az.shape, dtype=np.double)
        if return_pa:
            pa = check_output('pa', pa, shape=az.shape, dtype=np.double)
        else:
            sin2psi = check_output('sin2psi', sin2psi, shape=az.shape, dtype=np.double)
            cos2psi = check_output('cos2psi', cos2psi, shape=az.shape, dtype=np.double)
        n = az.size

        if hwp is None:
            if return_pa:
                qp.qp_azelpsi2radecpa(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    psi,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    ra,
                    dec,
                    pa,
                    n,
                )
            elif sindec:
                qp.qp_azelpsi2rasindec(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    psi,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
            else:
                qp.qp_azelpsi2radec(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    psi,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
        else:
            hwp = check_input('hwp', hwp, shape=az.shape)

            if return_pa:
                qp.qp_azelpsi2radec_hwp(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    psi,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    hwp,
                    ra,
                    dec,
                    pa,
                    n,
                )
            elif sindec:
                qp.qp_azelpsi2rasindec_hwp(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    psi,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    hwp,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )
            else:
                qp.qp_azelpsi2radec_hwp(
                    self._memory,
                    delta_az,
                    delta_el,
                    delta_psi,
                    az,
                    el,
                    psi,
                    pitch,
                    roll,
                    lon,
                    lat,
                    ctime,
                    hwp,
                    ra,
                    dec,
                    sin2psi,
                    cos2psi,
                    n,
                )

        if return_pa:
            return ra, dec, pa
        return ra, dec, sin2psi, cos2psi

    def radec2azel(
        self, ra, dec, pa, lon, lat, ctime, az=None, el=None, hpa=None, **kwargs
    ):
        """
        Estimate the horizon coordinates for a given set of equatorial coordinates
        (ra/dec/psi), location on the earth (lon/lat) and UTC time.  Input vectors
        must be numpy-array-like and broadcastable to the same shape.

        Arguments
        ---------
        ra : array_like
            Right ascension angle
        dec : array_like
            Declination angle
        pa : array_like
            Position angle in equatorial coordinates
        lon : array_like
            Observer longitude in degrees.
        lat : array_like
            Observer latitude in degrees.
        ctime : array_like
            Unix time in seconds UTC

        Returns
        -------
        az : array_like
            Azimuth in degrees
        el : array_like
            Elevation in degrees
        hpa : array_like
            Position angle in horizon coordinates

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        ra, dec, pa, lon, lat, ctime = check_inputs(ra, dec, pa, lon, lat, ctime)

        az = check_output('az', az, shape=ra.shape, dtype=np.double)
        el = check_output('el', el, shape=ra.shape, dtype=np.double)
        hpa = check_output('hpa', hpa, shape=ra.shape, dtype=np.double)
        n = ra.size

        qp.qp_radec2azel(self._memory, ra, dec, pa, lon, lat, ctime, az, el, hpa, n)

        return az, el, hpa

    def radecpa2quat(self, ra, dec, pa, **kwargs):
        """
        Calculate quaternion for input ra/dec/pa. Vectorized, input arguments
        must be broadcastable to the same shape.

        Arguments
        ---------
        ra : array_like
            Right ascension angle
        dec : array_like
            Declination angle
        pa : array_like
            Position angle

        Returns
        -------
        q : array_like
            Quaternion constructed from the input angles.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """
        self.set(**kwargs)

        ra, dec, pa = check_inputs(ra, dec, pa)
        n = ra.size
        quat = check_output('quat', shape=(n, 4), dtype=np.double, **kwargs)
        qp.qp_radecpa2quatn(self._memory, ra, dec, pa, quat, n)

        if n == 1:
            return quat[0]
        return quat

    def quat2radecpa(self, quat, **kwargs):
        """
        Calculate ra/dec/pa for input quaternion(s).

        Arguments
        ---------
        q : array_like
            Pointing quaternion

        Returns
        -------
        ra : array_like
            Right ascension angle
        dec : array_like
            Declination angle
        pa : array_like
            Position angle

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """
        self.set(**kwargs)

        quat = check_input('quat', np.atleast_2d(quat), quat=True)
        n = quat.shape[0]
        ra = check_output('ra', shape=(n,), dtype=np.double, **kwargs)
        dec = check_output('dec', shape=(n,), dtype=np.double, **kwargs)
        pa = check_output('pa', shape=(n,), dtype=np.double, **kwargs)

        qp.qp_quat2radecpan(self._memory, quat, ra, dec, pa, n)
        if n == 1:
            return ra[0], dec[0], pa[0]
        return ra, dec, pa

    def quat2pixpa(self, quat, nside=256, **kwargs):
        """
        Calculate ra/dec/pa for input quaternion(s).

        Arguments
        ---------
        q : array_like
            Pointing quaternion

        Returns
        -------
        pix : array_like
            Pixel number(s) corresponding to the input positions(s).
        pa : array_like
            Position angle

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """
        self.set(**kwargs)

        quat = check_input('quat', np.atleast_2d(quat), quat=True)
        n = quat.shape[0]
        pix = check_output('pix', shape=(n,), dtype=int, **kwargs)
        pa = check_output('pa', shape=(n,), dtype=np.double, **kwargs)

        qp.qp_quat2pixpan(self._memory, quat, pix, pa, n)
        if n == 1:
            return pix[0], pa[0]
        return pix, pa

    def radec2pix(self, ra, dec, nside=256, **kwargs):
        """
        Calculate HEALpix pixel number for given ra/dec and nside.  Vectorized,
        input arguments must be broadcastable to the same shape.

        Arguments
        ---------
        ra : array_like
            Right ascension angle
        dec : array_like
            Declination angle
        nside : int
            HEALpix resolution parameter

        Returns
        -------
        pix : array_like
            Pixel number(s) corresponding to the input positions(s).

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        ra, dec = check_inputs(ra, dec)
        n = ra.size

        pix = check_output('pix', shape=ra.shape, dtype=int, **kwargs)
        qp.qp_radec2pixn(self._memory, ra, dec, nside, pix, n)

        if n == 1:
            return pix[0]
        return pix

    def rotate_quat(self, quat, coord=['C', 'G'], inplace=True, **kwargs):
        """
        Rotate a quaternion from one coordinate system to another.
        Supported coordinates:

        C: celestial (equatorial) coordinates
        G: galactic coordinates

        Arguments
        ---------
        quat : array_like
            array of quaternions, of shape (n, 4)
        coord : list, optional
            2-element list of input and output coordinates
        inplace : bool, optional
            If True, apply the rotation in-place on the input quaternion.
            Otherwise, return a copy of the input array.  Default: True.

        Returns
        -------
        quat : array_like
            rotated quaternion array

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        quat = check_input(
            'quat', np.atleast_2d(quat), quat=True, inplace=inplace, output=True
        )
        n = quat.size // 4

        if coord[0] == 'C' and coord[1] == 'G':
            qp.qp_radec2gal_quatn(self._memory, quat, n)
        elif coord[0] == 'G' and coord[1] == 'C':
            qp.qp_gal2radec_quatn(self._memory, quat, n)
        else:
            raise ValueError('unsupported coord {}'.format(coord))

        return quat.squeeze()

    def rotate_coord(
        self,
        ra,
        dec,
        pa=None,
        sin2psi=None,
        cos2psi=None,
        coord=['C', 'G'],
        inplace=True,
        **kwargs
    ):
        """
        Rotate coordinates from one coordinate system to another. Vectorized,
        input arguments must be broadcastable to the same shape.
        Supported coordinates:

        C: celestial (equatorial) coordinates
        G: galactic coordinates

        Arguments
        ---------
        ra, dec, pa : array_like
          -- or --
        ra, dec, sin2psi, cos2psi : array_like
            arrays of coordinates, of shape (n,)
            If none of pa, sin2psi or cos2psi are supplied,
            pa = 0 is assumed.
        coord : list, optional
            2-element list of input and output coordinates.
            Supported systems: C, G.
        inplace : bool, optional
            If True, apply the rotation in-place on the input coordinates.
            Otherwise, return a copy of the input array.  Default: True.

        Returns
        -------
        ra, dec, pa : array_like
          -- or --
        ra, dec, sin2psi, cos2psi : array_like
            rotated coordinate arrays, same form as input

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """
        self.set(**kwargs)

        do_pa = True
        if pa is None:
            if sin2psi is None and cos2psi is None:
                pass
            elif sin2psi is None or cos2psi is None:
                raise KeyError('both sin2psi and cos2psi arguments required')
            else:
                do_pa = False
        else:
            if sin2psi is not None or cos2psi is not None:
                raise KeyError(
                    'ambiguous pol arguments, supply either pa '
                    'or sin2psi/cos2psi only'
                )

        ra, dec, pa, sin2psi, cos2psi = check_inputs(
            ra, dec, pa, sin2psi, cos2psi, inplace=inplace, output=True
        )
        n = ra.size

        if coord[0] == 'C' and coord[1] == 'G':
            if do_pa:
                qp.qp_radecpa2galn(self._memory, ra, dec, pa, n)
            else:
                qp.qp_radec2galn(self._memory, ra, dec, sin2psi, cos2psi, n)
        elif coord[0] == 'G' and coord[1] == 'C':
            if do_pa:
                qp.qp_gal2radecpan(self._memory, ra, dec, pa, n)
            else:
                qp.qp_gal2radecn(self._memory, ra, dec, sin2psi, cos2psi, n)

        if n == 1:
            if do_pa:
                return ra[0], dec[0], pa[0]
            return ra[0], dec[0], sin2psi[0], cos2psi[0]
        if do_pa:
            return ra, dec, pa
        return ra, dec, sin2psi, cos2psi

    def radec2gal(
        self, ra, dec, pa=None, sin2psi=None, cos2psi=None, inplace=True, **kwargs
    ):
        """
        Rotate celestial (equatorial) coordinates to galactic coordinates.  This
        is equivalent to calling :meth:`rotate_coord` with `coord=['C', 'G']`.
        Vectorized, input arguments must be broadcastable to the same shape.

        Arguments
        ---------
        ra, dec, pa : array_like
          -- or --
        ra, dec, sin2psi, cos2psi : array_like
            arrays of coordinates, of shape (n,).  If none of pa, sin2psi or
            cos2psi are supplied, pa = 0 is assumed.
        inplace : bool, optional
            If True, apply the rotation in-place on the input quaternion.
            Otherwise, return a copy of the input array.  Default: True.

        Returns
        -------
        l, b, pa : array_like
          -- or --
        l, b, sin2psi, cos2psi : array_like
            rotated coordinate arrays, same form as input

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """
        return self.rotate_coord(
            ra, dec, pa, sin2psi, cos2psi, coord=['C', 'G'], inplace=inplace, **kwargs
        )

    def gal2radec(
        self, ra, dec, pa=None, sin2psi=None, cos2psi=None, inplace=True, **kwargs
    ):
        """
        Rotate galactic coordinates to celestial (equatorial) coordinates.  This
        is equivalent to calling :meth:`rotate_coord` with `coord=['G', 'C']`.
        Vectorized, input arguments must be broadcastable to the same shape.

        Arguments
        ---------
        l, b, pa : array_like
          -- or --
        l, b, sin2psi, cos2psi : array_like
            arrays of coordinates, of shape (n,).  If none of pa, sin2psi or
            cos2psi are supplied, pa = 0 is assumed.
        inplace : bool, optional
            If True, apply the rotation in-place on the input quaternion.
            Otherwise, return a copy of the input array.  Default: True.

        Returns
        -------
        ra, dec, pa : array_like
          -- or --
        ra, dec, sin2psi, cos2psi : array_like
            rotated coordinate arrays, same form as input

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """
        return self.rotate_coord(
            ra, dec, pa, sin2psi, cos2psi, coord=['G', 'C'], inplace=inplace, **kwargs
        )

    def rotate_map(
        self, map_in, coord=['C', 'G'], map_out=None, interp_pix=True, **kwargs
    ):
        """
        Rotate a polarized 3-x-npix map from one coordinate system to another.
        Supported coordinates:

        C = celestial (equatorial J2000)
        G = galactic

        Arguments
        ---------
        map_in : array_like
            Input map, of shape (3, N)
        coord : list, optional
            2-element list of input and output coordinates.
            Supported systems: C, G.
        map_out : array_like, optional
            Rotated output map, for inplace operation.  Same shape as `map_in`.
        interp_pix : bool, optional
            If True, interpolate the rotated map.

        Returns
        -------
        map_out : array_like
            Rotated output map.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.

        Only full-sky maps are currently supported.
        """

        warn('This code is buggy, use at your own risk', UserWarning)

        interp_orig = self.get('interp_pix')
        self.set(interp_pix=interp_pix, **kwargs)

        from .qmap_class import check_map

        map_in, nside = check_map(map_in)
        map_out = check_output(
            'map_out', map_out, shape=map_in.shape, dtype=map_in.dtype, fill=0
        )

        try:
            coord_in = coord[0]
            coord_out = coord[1]
        except:
            raise ValueError('unable to parse coord')

        map_in_p = lib.pointer_2d(map_in)
        map_out_p = lib.pointer_2d(map_out)

        qp.qp_rotate_map(self._memory, nside, map_in_p, coord_in, map_out_p, coord_out)

        self.set(interp_pix=interp_orig)
        return map_out

    def quat2pix(self, quat, nside=256, pol=True, **kwargs):
        """
        Calculate HEALpix pixel number and optional polarization angle
        for a given orientation.

        Arguments
        ---------
        quat : quaternion or array of quaternions
            Pointing orientation(s)
        nside : int, optional
            HEALpix resolution parameter
        pol : bool, optional
            If True, return sin2psi and cos2psi along with the pixel number(s)

        Returns
        -------
        pix : array_like
            Pixel number(s) for the given input quaternion(s)
        sin2psi : array_like
        cos2psi : array_like
            Polarization coefficients, if `pol` is `True`.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        quat = check_input('quat', np.atleast_2d(quat), quat=True)

        n = quat.shape[0]
        shape = (n,)
        pix = check_output('pix', shape=shape, dtype=int, **kwargs)
        sin2psi = check_output('sin2psi', shape=shape, **kwargs)
        cos2psi = check_output('cos2psi', shape=shape, **kwargs)
        qp.qp_quat2pixn(self._memory, quat, nside, pix, sin2psi, cos2psi, n)

        if n == 1:
            pix, sin2psi, cos2psi = pix[0], sin2psi[0], cos2psi[0]
        if pol:
            return pix, sin2psi, cos2psi
        return pix

    def bore2pix(
        self,
        q_off,
        ctime,
        q_bore,
        q_hwp=None,
        nside=256,
        pol=True,
        return_pa=False,
        **kwargs
    ):
        """
        Calculate the orientation on the sky for a detector offset from the
        boresight.  Detector offsets are defined assuming the boresight is
        pointed toward the horizon, and that the boresight polarization axis is
        along the vertical.

        Arguments
        ---------
        q_off : quaternion
            Detector offset quaternion for a single detector,
            calculated using :meth:`det_offset`.
        ctime : array_like
            Unix times in seconds UTC, broadcastable to shape (N,),
            the long dimenions of `q_bore`.
        q_bore : quaternion or array of quaternions
            Nx4 array of quaternions encoding the boresight orientation on the
            sky (as output by :meth:`azel2radec` or equivalent)
        q_hwp : quaternion or array of quaternions, optional
            HWP angle quaternions calculated using :meth:`hwp_quat`.  Must be
            broadcastable to the same shape as `q_bore`.
        nside : int, optional
            HEALpix map dimension.  Default: 256.
        pol : bool, optional
            If `False`, return only the pixel timestream
        return_pa : bool, optional
            If `True`, return pa instead of sin2psi / cos2psi

        Returns
        -------
        pix : array_like
            Detector pixel number
        pa/sin2psi : array_like
            Detector polarization orientation if `return_pa` is `True`, or
            sin(2*pa) if `return_pa` is `False`.
        cos2psi : array_like
            detector polarization orientation cos(2*pa), if `return_pa` is `False`.

        Notes
        -----
        Any keywords accepted by the :meth:`set` method can also be passed here,
        and will be processed prior to calculation.
        """

        self.set(**kwargs)

        q_off = check_input('q_off', q_off, quat=True)
        q_bore = check_input('q_bore', q_bore, quat=True)
        if ctime is None:
            if not self.get('mean_aber'):
                raise ValueError('ctime required if mean_aber is False')
            ctime = np.zeros((q_bore.size // 4,), dtype=q_bore.dtype)
        ctime = check_input('ctime', ctime)
        pix = check_output('pix', shape=ctime.shape, dtype=int, **kwargs)
        if return_pa:
            pa = check_output('pa', shape=ctime.shape, **kwargs)
        else:
            sin2psi = check_output('sin2psi', shape=ctime.shape, **kwargs)
            cos2psi = check_output('cos2psi', shape=ctime.shape, **kwargs)
        n = ctime.size

        if q_hwp is None:
            if return_pa:
                qp.qp_bore2pixpa(self._memory, q_off, ctime, q_bore, nside, pix, pa, n)
            else:
                qp.qp_bore2pix(
                    self._memory, q_off, ctime, q_bore, nside, pix, sin2psi, cos2psi, n
                )
        else:
            q_hwp = check_input('q_hwp', q_hwp, shape=q_bore.shape)

            if return_pa:
                qp.qp_bore2pixpa_hwp(
                    self._memory, q_off, ctime, q_bore, q_hwp, nside, pix, pa, n
                )
            else:
                qp.qp_bore2pix_hwp(
                    self._memory,
                    q_off,
                    ctime,
                    q_bore,
                    q_hwp,
                    nside,
                    pix,
                    sin2psi,
                    cos2psi,
                    n,
                )

        if pol is True:
            if return_pa:
                return pix, pa
            return pix, sin2psi, cos2psi
        return pix

    def get_interp_val(self, map_in, ra, dec, nest=False):
        """
        Interpolate map pixels to these coordinates.  Uses a C implementation
        of the bilinear interpolation method `get_interpol()` as implemented
        in the equivalent healpix_cxx / healpy function.

        Arguments
        ---------
        map_in : array_like
            A single healpix map or list of maps which to interpolate from.
        ra, dec: array_like
            Timestreams of coordinates to interpolate to, in degrees,
            broadcastable to a common shape (nsample,)
        nest: bool, optional
            If True, input map is in the nested pixel ordering scheme.
            Otherwise, ring ordering is assumed.
            Default: False.

        Returns
        -------
        values : array_like
            Array of interpolated map values, of shape (nmap, nsample).
        """

        pix_order = self.get('pix_order')
        if nest:
            self.set(pix_order='nest')
        else:
            self.set(pix_order='ring')

        ra, dec = check_inputs(ra, dec)
        n = ra.size

        from .qmap_class import check_map

        map_in, nside = check_map(map_in)

        val = check_output('value', shape=(len(map_in), n))

        for m, v in zip(map_in, val):
            qp.qp_get_interp_valn(self._memory, nside, m, ra, dec, v, n)

        self.set(pix_order=pix_order)

        v = v.squeeze()
        if not v.shape:
            return v[()]
        return v

    def update_bulletin_a(self, start_year=2000):
        """
        Update the IERS Bulletin A database using astropy, and return the stored
        entries.  Issues an ImportWarning if astropy version 1.2 or newer is not
        found.

        Arguments
        ---------
        start_year : int, optional
            Oldest year for which data should be stored.

        Returns
        -------
        mjd : array_like
            Modified Julian date
        dut1 : array_like
            UT1-UTC time correction
        x : array_like
        y : array_like
            Polar motion (wobble) corrections
        """
        try:
            from astropy.utils.iers import IERS_Auto
        except ImportError:
            warn(
                'Compatible Astropy not found.  Install astropy v1.2 or newer '
                'for accurate polar motion and UT1 corrections',
                ImportWarning,
            )
            return

        columns = ['MJD', 'UT1_UTC', 'PM_x', 'PM_y', 'year']
        iers_table = IERS_Auto.open()[columns].as_array()

        # check year
        year = iers_table['year'] + 1900
        (wraps,) = np.where(np.ediff1d(year) < 0)
        for idx in wraps:
            year[idx + 1 :] += 100
        iers_table['year'] = year
        iers_table = iers_table[year >= start_year]

        # check MJD
        mjds = iers_table['MJD']
        mjd_min = int(mjds.min())
        mjd_max = int(mjds.max())

        # update table
        dut1 = np.array(iers_table['UT1_UTC'])
        x = np.array(iers_table['PM_x'])
        y = np.array(iers_table['PM_y'])
        qp.qp_set_iers_bulletin_a(self._memory, mjd_min, mjd_max, dut1, x, y)

        return mjds, dut1, x, y

    def load_bulletin_a(self, filename, columns=['mjd', 'dut1', 'x', 'y'], **kwargs):
        """
        Load IERS Bulletin A from file and store in memory.  The file must be
        readable using `numpy.loadtxt` with `unpack=True`, and is assumed to be
        sorted by mjd.

        Arguments
        ---------
        filename : string
            Name of the text file containing IERS Bulletin A parameters.
        columns : list of strings
            list of columns as they appear in the file.
            A KeyError is raise if the list does not contain
            each of ['mjd', 'dut1', 'x', 'y'].

        Any other keyword arguments are passed to the `numpy.loadtxt` function

        Returns
        -------
        mjd : array_like
            Modified Julian date
        dut1 : array_like
            UT1-UTC time correction
        x : array_like
        y : array_like
            Polar motion corrections
        """

        req_columns = ['mjd', 'dut1', 'x', 'y']
        if not set(req_columns) <= set(columns):
            raise KeyError(
                'Missing columns {}'.format(list(set(req_columns) - set(columns)))
            )
        kwargs['unpack'] = True
        data = np.loadtxt(filename, **kwargs)
        mjd, x, y, dut1 = (data[columns.index(x)] for x in req_columns)
        mjd_min, mjd_max = int(mjd[0]), int(mjd[-1])

        try:
            qp.qp_set_iers_bulletin_a(self._memory, mjd_min, mjd_max, dut1, x, y)
        except:
            raise RuntimeError(
                'Error loading Bulletin A data from file {}'.format(filename)
            )

        return mjd, dut1, x, y

    def get_bulletin_a(self, mjd):
        """
        Return dut1/x/y for given mjd. Numpy-vectorized.
        """

        def func(x):
            return lib.qp_get_bulletin_a(self._memory, x)

        fvec = np.vectorize(func, [np.double] * 3)

        out = fvec(mjd)
        if out[0].shape == ():
            return tuple(x[()] for x in out)
        return out



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from ._libqpoint import libqp as qp

__all__ = ['refraction']


def refraction(el, temp, press, hum, freq=150.0):
    """
    Standalone function for calculating the refraction correction without
    storing any parameters.  Useful for testing, numpy-vectorized.

    Arguments
    ---------
    el : array_like
        Observer elevation angle, degrees
    temperature : array_like
        Ambient temperature, Celcius
    pressure : array_like
        Ambient pressure, mbar
    humidity : array_like
        Relative humidity, fraction
    frequency : array_like
        Observing frequency, GHz

    Returns
    -------
    delta : array_like
        Refraction correction, in degrees
    """

    fvec = np.vectorize(qp.qp_refraction, [np.double])
    delta = fvec(el, temp, press, hum, freq)
    if delta.shape == ():
        return delta[()]
    return delta
